# Copyright 2020 The TensorFlow Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for vector."""

from absl.testing import flagsaver
from absl.testing import parameterized
import numpy as np
import tensorflow as tf

from tensorflow_graphics.geometry.transformation.tests import test_data as td
from tensorflow_graphics.math import vector
from tensorflow_graphics.util import test_case


class VectorTest(test_case.TestCase):

  @parameterized.parameters(
      ((None, 3), (None, 3)),)
  def test_cross_exception_not_raised(self, *shapes):
    """Tests that the shape exceptions are not raised."""
    self.assert_exception_is_not_raised(vector.cross, shapes)

  @parameterized.parameters(
      ("must have exactly 3 dimensions in axis", (1,), (3,)),
      ("must have exactly 3 dimensions in axis", (3,), (2,)),
      ("Not all batch dimensions are broadcast-compatible.", (2, 3), (3, 3)),
  )
  def test_cross_exception_raised(self, error_msg, *shapes):
    """Tests that the shape exceptions are properly raised."""
    self.assert_exception_is_raised(vector.cross, error_msg, shapes)

  @parameterized.parameters(
      (td.AXIS_3D_0, td.AXIS_3D_0),
      (td.AXIS_3D_0, td.AXIS_3D_X),
      (td.AXIS_3D_0, td.AXIS_3D_Y),
      (td.AXIS_3D_0, td.AXIS_3D_Z),
      (td.AXIS_3D_X, td.AXIS_3D_X),
      (td.AXIS_3D_X, td.AXIS_3D_Y),
      (td.AXIS_3D_X, td.AXIS_3D_Z),
      (td.AXIS_3D_Y, td.AXIS_3D_X),
      (td.AXIS_3D_Y, td.AXIS_3D_Y),
      (td.AXIS_3D_Y, td.AXIS_3D_Z),
      (td.AXIS_3D_Z, td.AXIS_3D_X),
      (td.AXIS_3D_Z, td.AXIS_3D_Y),
      (td.AXIS_3D_Z, td.AXIS_3D_Z),
  )
  def test_cross_jacobian_preset(self, u_init, v_init):
    """Tests the Jacobian of the dot product."""
    self.assert_jacobian_is_correct_fn(vector.cross, [u_init, v_init])

  def test_cross_jacobian_random(self):
    """Test the Jacobian of the dot product."""
    tensor_size = np.random.randint(3)
    tensor_shape = np.random.randint(1, 10, size=(tensor_size)).tolist()
    u_init = np.random.random(size=tensor_shape + [3])
    v_init = np.random.random(size=tensor_shape + [3])

    self.assert_jacobian_is_correct_fn(vector.cross, [u_init, v_init])

  @parameterized.parameters(
      ((td.AXIS_3D_0, td.AXIS_3D_0), (td.AXIS_3D_0,)),
      ((td.AXIS_3D_0, td.AXIS_3D_X), (td.AXIS_3D_0,)),
      ((td.AXIS_3D_0, td.AXIS_3D_Y), (td.AXIS_3D_0,)),
      ((td.AXIS_3D_0, td.AXIS_3D_Z), (td.AXIS_3D_0,)),
      ((td.AXIS_3D_X, td.AXIS_3D_X), (td.AXIS_3D_0,)),
      ((td.AXIS_3D_X, td.AXIS_3D_Y), (td.AXIS_3D_Z,)),
      ((td.AXIS_3D_X, td.AXIS_3D_Z), (-td.AXIS_3D_Y,)),
      ((td.AXIS_3D_Y, td.AXIS_3D_X), (-td.AXIS_3D_Z,)),
      ((td.AXIS_3D_Y, td.AXIS_3D_Y), (td.AXIS_3D_0,)),
      ((td.AXIS_3D_Y, td.AXIS_3D_Z), (td.AXIS_3D_X,)),
      ((td.AXIS_3D_Z, td.AXIS_3D_X), (td.AXIS_3D_Y,)),
      ((td.AXIS_3D_Z, td.AXIS_3D_Y), (-td.AXIS_3D_X,)),
      ((td.AXIS_3D_Z, td.AXIS_3D_Z), (td.AXIS_3D_0,)),
  )
  def test_cross_preset(self, test_inputs, test_outputs):
    """Tests the cross product of predefined axes."""
    self.assert_output_is_correct(vector.cross, test_inputs, test_outputs)

  def test_cross_random(self):
    """Tests the cross product function."""
    tensor_size = np.random.randint(1, 4)
    tensor_shape = np.random.randint(1, 10, size=tensor_size).tolist()
    axis = np.random.randint(tensor_size)
    tensor_shape[axis] = 3  # pylint: disable=invalid-sequence-index
    u = np.random.random(size=tensor_shape)
    v = np.random.random(size=tensor_shape)

    self.assertAllClose(
        vector.cross(u, v, axis=axis), np.cross(u, v, axis=axis))

  @parameterized.parameters(
      ((None,), (None,)),
      ((None, None), (None, None)),
  )
  def test_dot_exception_not_raised(self, *shapes):
    """Tests that the shape exceptions are not raised."""
    self.assert_exception_is_not_raised(vector.dot, shapes)

  @parameterized.parameters(
      ("must have the same number of dimensions", (None, 1), (None, 2)),
      ("Not all batch dimensions are broadcast-compatible.", (2, 3), (3, 3)),
  )
  def test_dot_exception_raised(self, error_msg, *shapes):
    """Tests that the shape exceptions are properly raised."""
    self.assert_exception_is_raised(vector.dot, error_msg, shapes)

  @parameterized.parameters(
      (td.AXIS_3D_0, td.AXIS_3D_0),
      (td.AXIS_3D_0, td.AXIS_3D_X),
      (td.AXIS_3D_0, td.AXIS_3D_Y),
      (td.AXIS_3D_0, td.AXIS_3D_Z),
      (td.AXIS_3D_X, td.AXIS_3D_X),
      (td.AXIS_3D_X, td.AXIS_3D_Y),
      (td.AXIS_3D_X, td.AXIS_3D_Z),
      (td.AXIS_3D_Y, td.AXIS_3D_X),
      (td.AXIS_3D_Y, td.AXIS_3D_Y),
      (td.AXIS_3D_Y, td.AXIS_3D_Z),
      (td.AXIS_3D_Z, td.AXIS_3D_X),
      (td.AXIS_3D_Z, td.AXIS_3D_Y),
      (td.AXIS_3D_Z, td.AXIS_3D_Z),
  )
  def test_dot_jacobian_preset(self, u_init, v_init):
    """Tests the Jacobian of the dot product."""
    self.assert_jacobian_is_correct_fn(vector.dot, [u_init, v_init])

  def test_dot_jacobian_random(self):
    """Tests the Jacobian of the dot product."""
    tensor_size = np.random.randint(3)
    tensor_shape = np.random.randint(1, 10, size=(tensor_size)).tolist()
    u_init = np.random.random(size=tensor_shape + [3])
    v_init = np.random.random(size=tensor_shape + [3])

    self.assert_jacobian_is_correct_fn(vector.dot, [u_init, v_init])

  @parameterized.parameters(
      ((td.AXIS_3D_0, td.AXIS_3D_0), (0.,)),
      ((td.AXIS_3D_0, td.AXIS_3D_X), (0.,)),
      ((td.AXIS_3D_0, td.AXIS_3D_Y), (0.,)),
      ((td.AXIS_3D_0, td.AXIS_3D_Z), (0.,)),
      ((td.AXIS_3D_X, td.AXIS_3D_X), (1.,)),
      ((td.AXIS_3D_X, td.AXIS_3D_Y), (0.,)),
      ((td.AXIS_3D_X, td.AXIS_3D_Z), (0.,)),
      ((td.AXIS_3D_Y, td.AXIS_3D_X), (0.,)),
      ((td.AXIS_3D_Y, td.AXIS_3D_Y), (1.,)),
      ((td.AXIS_3D_Y, td.AXIS_3D_Z), (0.,)),
      ((td.AXIS_3D_Z, td.AXIS_3D_X), (0.,)),
      ((td.AXIS_3D_Z, td.AXIS_3D_Y), (0.,)),
      ((td.AXIS_3D_Z, td.AXIS_3D_Z), (1.,)),
  )
  def test_dot_preset(self, test_inputs, test_outputs):
    """Tests the dot product of predefined axes."""

    def func(u, v):
      return tf.squeeze(vector.dot(u, v), axis=-1)

    self.assert_output_is_correct(func, test_inputs, test_outputs)

  def test_dot_random(self):
    """Tests the dot product function."""
    tensor_size = np.random.randint(2, 4)
    tensor_shape = np.random.randint(1, 10, size=tensor_size).tolist()
    axis = np.random.randint(tensor_size)
    u = np.random.random(size=tensor_shape)
    v = np.random.random(size=tensor_shape)

    dot = tf.linalg.tensor_diag_part(tf.tensordot(u, v, axes=[[axis], [axis]]))
    dot = tf.expand_dims(dot, axis=axis)

    self.assertAllClose(vector.dot(u, v, axis=axis), dot)

  @parameterized.parameters(
      ((None,), (None,)),
      ((None, None), (None, None)),
      ((1,), (1,)),
      ((1, 1), (1, 1)),
  )
  def test_reflect_exception_not_raised(self, *shapes):
    """Tests that the shape exceptions are not raised."""
    self.assert_exception_is_not_raised(vector.reflect, shapes)

  @parameterized.parameters(
      ("must have the same number of dimensions", (None, 1), (None, 2)),
      ("Not all batch dimensions are broadcast-compatible.", (2, 2), (3, 2)),
  )
  def test_reflect_exception_raised(self, error_msg, *shapes):
    """Tests that the shape exceptions are properly raised."""
    self.assert_exception_is_raised(vector.reflect, error_msg, shapes)

  @parameterized.parameters(
      (td.AXIS_3D_0, td.AXIS_3D_0),
      (td.AXIS_3D_0, td.AXIS_3D_X),
      (td.AXIS_3D_0, td.AXIS_3D_Y),
      (td.AXIS_3D_0, td.AXIS_3D_Z),
      (td.AXIS_3D_X, td.AXIS_3D_X),
      (td.AXIS_3D_X, td.AXIS_3D_Y),
      (td.AXIS_3D_X, td.AXIS_3D_Z),
      (td.AXIS_3D_Y, td.AXIS_3D_X),
      (td.AXIS_3D_Y, td.AXIS_3D_Y),
      (td.AXIS_3D_Y, td.AXIS_3D_Z),
      (td.AXIS_3D_Z, td.AXIS_3D_X),
      (td.AXIS_3D_Z, td.AXIS_3D_Y),
      (td.AXIS_3D_Z, td.AXIS_3D_Z),
  )
  @flagsaver.flagsaver(tfg_add_asserts_to_graph=False)
  def test_reflect_jacobian_preset(self, u_init, v_init):
    """Tests the Jacobian of the reflect function."""
    self.assert_jacobian_is_correct_fn(vector.reflect, [u_init, v_init])

  @flagsaver.flagsaver(tfg_add_asserts_to_graph=False)
  def test_reflect_jacobian_random(self):
    """Tests the Jacobian of the reflect function."""
    tensor_size = np.random.randint(3)
    tensor_shape = np.random.randint(1, 10, size=(tensor_size)).tolist()
    u_init = np.random.random(size=tensor_shape + [3])
    v_init = np.random.random(size=tensor_shape + [3])

    self.assert_jacobian_is_correct_fn(vector.reflect, [u_init, v_init])

  @parameterized.parameters(
      ((td.AXIS_3D_0, td.AXIS_3D_X), (td.AXIS_3D_0,)),
      ((td.AXIS_3D_0, td.AXIS_3D_Y), (td.AXIS_3D_0,)),
      ((td.AXIS_3D_0, td.AXIS_3D_Z), (td.AXIS_3D_0,)),
      ((td.AXIS_3D_X, td.AXIS_3D_X), (-td.AXIS_3D_X,)),
      ((td.AXIS_3D_X, td.AXIS_3D_Y), (td.AXIS_3D_X,)),
      ((td.AXIS_3D_X, td.AXIS_3D_Z), (td.AXIS_3D_X,)),
      ((td.AXIS_3D_Y, td.AXIS_3D_X), (td.AXIS_3D_Y,)),
      ((td.AXIS_3D_Y, td.AXIS_3D_Y), (-td.AXIS_3D_Y,)),
      ((td.AXIS_3D_Y, td.AXIS_3D_Z), (td.AXIS_3D_Y,)),
      ((td.AXIS_3D_Z, td.AXIS_3D_X), (td.AXIS_3D_Z,)),
      ((td.AXIS_3D_Z, td.AXIS_3D_Y), (td.AXIS_3D_Z,)),
      ((td.AXIS_3D_Z, td.AXIS_3D_Z), (-td.AXIS_3D_Z,)),
  )
  def test_reflect_preset(self, test_inputs, test_outputs):
    """Tests the reflect function of predefined axes."""
    self.assert_output_is_correct(vector.reflect, test_inputs, test_outputs)

  def test_reflect_random(self):
    """Tests that calling reflect twice give an identity transform."""
    tensor_size = np.random.randint(2, 4)
    tensor_shape = np.random.randint(2, 3, size=tensor_size).tolist()
    axis = np.random.randint(tensor_size)
    u = np.random.random(size=tensor_shape)
    v = np.random.random(size=tensor_shape)
    v /= np.linalg.norm(v, axis=axis, keepdims=True)

    u_new = vector.reflect(u, v, axis=axis)
    u_new = vector.reflect(u_new, v, axis=axis)

    self.assertAllClose(u_new, u)


if __name__ == "__main__":
  test_case.main()



# Copyright 2020 The TensorFlow Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
r"""This module implements perspective camera functionalities.

The perspective camera model, also referred to as pinhole camera model, is
defined using a focal length \\((f_x, f_y)\\) and a principal point
\\((c_x, c_y)\\). The perspective camera model can be written as a calibration
matrix

$$
\mathbf{C} =
\begin{bmatrix}
f_x & 0 & c_x \\
0  & f_y & c_y \\
0  & 0  & 1 \\
\end{bmatrix},
$$

also referred to as the intrinsic parameter matrix. The camera focal length
\\((f_x, f_y)\\), defined in pixels, is the physical focal length divided by the
physical size of a camera pixel. The physical focal length is the distance
between the camera center and the image plane. The principal point is the
intersection of the camera axis with the image plane. The camera axis is the
line perpendicular to the image plane starting at the optical center.

More details about perspective cameras can be found on [this page.]
(http://ksimek.github.io/2013/08/13/intrinsic/)

Note: The current implementation does not take into account distortion or
skew parameters.
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import math
import tensorflow as tf

from tensorflow_graphics.util import asserts
from tensorflow_graphics.util import export_api
from tensorflow_graphics.util import safe_ops
from tensorflow_graphics.util import shape


def parameters_from_right_handed(projection_matrix, name=None):
  """Recovers the parameters used to contruct a right handed projection matrix.

  Note:
    In the following, A1 to An are optional batch dimensions.

  Args:
    projection_matrix: A tensor of shape `[A1, ..., An, 4, 4]`, containing
      matrices of right handed perspective-view frustum.
    name: A name for this op. Defaults to
      'perspective_parameters_from_right_handed'.

  Raises:
    InvalidArgumentError: if `projection_matrix` is not of the expected shape.

  Returns:
    Tuple of 4 tensors of shape `[A1, ..., An, 1]`, where the first tensor
    represents the vertical field of view used to contruct `projection_matrix,
    the second tensor represents the ascpect ratio used to construct
    `projection_matrix`, and the third and fourth parameters repectively
    represent the near and far clipping planes used to construct
    `projection_matrix`.
  """
  with tf.compat.v1.name_scope(name, "perspective_parameters_from_right_handed",
                               [projection_matrix]):
    projection_matrix = tf.convert_to_tensor(value=projection_matrix)

    shape.check_static(
        tensor=projection_matrix,
        tensor_name="projection_matrix",
        has_rank_greater_than=1,
        has_dim_equals=((-2, 4), (-1, 4)))

    inverse_tan_half_vertical_field_of_view = projection_matrix[..., 1, 1:2]
    vertical_field_of_view = 2.0 * tf.atan(
        1.0 / inverse_tan_half_vertical_field_of_view)
    aspect_ratio = inverse_tan_half_vertical_field_of_view / projection_matrix[
        ..., 0, 0:1]

    a = projection_matrix[..., 2, 2:3]
    b = projection_matrix[..., 2, 3:4]

    far = b / (a + 1.0)
    near = (a + 1.0) / (a - 1.0) * far

    return vertical_field_of_view, aspect_ratio, near, far


def right_handed(vertical_field_of_view, aspect_ratio, near, far, name=None):
  """Generates the matrix for a right handed perspective projection.

  Note:
    In the following, A1 to An are optional batch dimensions.

  Args:
    vertical_field_of_view: A tensor of shape `[A1, ..., An, 1]`, where the last
      dimension represents the vertical field of view of the frustum expressed
      in radians. Note that values for `vertical_field_of_view` must be in the
      range (0,pi).
    aspect_ratio: A tensor of shape `[A1, ..., An, 1]`, where the last dimension
      stores the width over height ratio of the frustum. Note that values for
      `aspect_ratio` must be non-negative.
    near:  A tensor of shape `[A1, ..., An, 1]`, where the last dimension
      captures the distance between the viewer and the near clipping plane. Note
      that values for `near` must be non-negative.
    far:  A tensor of shape `[A1, ..., An, 1]`, where the last dimension
      captures the distance between the viewer and the far clipping plane. Note
      that values for `far` must be greater than those of `near`.
    name: A name for this op. Defaults to 'perspective_right_handed'.

  Raises:
    InvalidArgumentError: if any input contains data not in the specified range
      of valid values.
    ValueError: if the all the inputs are not of the same shape.

  Returns:
    A tensor of shape `[A1, ..., An, 4, 4]`, containing matrices of right
    handed perspective-view frustum.
  """
  with tf.compat.v1.name_scope(
      name, "perspective_right_handed",
      [vertical_field_of_view, aspect_ratio, near, far]):
    vertical_field_of_view = tf.convert_to_tensor(value=vertical_field_of_view)
    aspect_ratio = tf.convert_to_tensor(value=aspect_ratio)
    near = tf.convert_to_tensor(value=near)
    far = tf.convert_to_tensor(value=far)

    shape.check_static(
        tensor=vertical_field_of_view,
        tensor_name="vertical_field_of_view",
        has_dim_equals=(-1, 1))
    shape.check_static(
        tensor=aspect_ratio, tensor_name="aspect_ratio", has_dim_equals=(-1, 1))
    shape.check_static(tensor=near, tensor_name="near", has_dim_equals=(-1, 1))
    shape.check_static(tensor=far, tensor_name="far", has_dim_equals=(-1, 1))
    shape.compare_batch_dimensions(
        tensors=(vertical_field_of_view, aspect_ratio, near, far),
        last_axes=-2,
        tensor_names=("vertical_field_of_view", "aspect_ratio", "near", "far"),
        broadcast_compatible=False)

    vertical_field_of_view = asserts.assert_all_in_range(
        vertical_field_of_view, 0.0, math.pi, open_bounds=True)
    aspect_ratio = asserts.assert_all_above(aspect_ratio, 0.0, open_bound=True)
    near = asserts.assert_all_above(near, 0.0, open_bound=True)
    far = asserts.assert_all_above(far, near, open_bound=True)

    inverse_tan_half_vertical_field_of_view = 1.0 / tf.tan(
        vertical_field_of_view * 0.5)
    zero = tf.zeros_like(inverse_tan_half_vertical_field_of_view)
    one = tf.ones_like(inverse_tan_half_vertical_field_of_view)
    near_minus_far = near - far
    matrix = tf.concat(
        (inverse_tan_half_vertical_field_of_view / aspect_ratio, zero, zero,
         zero, zero, inverse_tan_half_vertical_field_of_view, zero, zero, zero,
         zero, (far + near) / near_minus_far, 2.0 * far * near / near_minus_far,
         zero, zero, -one, zero),
        axis=-1)
    matrix_shape = tf.shape(input=matrix)
    output_shape = tf.concat((matrix_shape[:-1], (4, 4)), axis=-1)
    return tf.reshape(matrix, shape=output_shape)


def intrinsics_from_matrix(matrix, name=None):
  r"""Extracts intrinsic parameters from a calibration matrix.

  Extracts the focal length \\((f_x, f_y)\\) and the principal point
  \\((c_x, c_y)\\) from a camera calibration matrix

  $$
  \mathbf{C} =
  \begin{bmatrix}
  f_x & 0 & c_x \\
  0  & f_y & c_y \\
  0  & 0  & 1 \\
  \end{bmatrix}.
  $$

  Note:
    In the following, A1 to An are optional batch dimensions.

  Args:
    matrix: A tensor of shape `[A1, ..., An, 3, 3]`, where the last two
      dimensions represent a camera calibration matrix.
    name: A name for this op that defaults to
      "perspective_intrinsics_from_matrix".

  Returns:
    Tuple of two tensors, each one of shape `[A1, ..., An, 2]`. The first
    tensor represents the focal length, and the second one the principle point.

  Raises:
    ValueError: If the shape of `matrix` is not supported.
  """
  with tf.compat.v1.name_scope(name, "perspective_intrinsics_from_matrix",
                               [matrix]):
    matrix = tf.convert_to_tensor(value=matrix)

    shape.check_static(
        tensor=matrix,
        tensor_name="matrix",
        has_rank_greater_than=1,
        has_dim_equals=((-1, 3), (-2, 3)))

    fx = matrix[..., 0, 0]
    fy = matrix[..., 1, 1]
    cx = matrix[..., 0, 2]
    cy = matrix[..., 1, 2]
    focal = tf.stack((fx, fy), axis=-1)
    principal_point = tf.stack((cx, cy), axis=-1)
  return focal, principal_point


def matrix_from_intrinsics(focal, principal_point, name=None):
  r"""Builds calibration matrix from intrinsic parameters.

  Builds the camera calibration matrix as

  $$
  \mathbf{C} =
  \begin{bmatrix}
  f_x & 0 & c_x \\
  0  & f_y & c_y \\
  0  & 0  & 1 \\
  \end{bmatrix}
  $$

  from the focal length \\((f_x, f_y)\\) and the principal point
  \\((c_x, c_y)\\).

  Note:
    In the following, A1 to An are optional batch dimensions.

  Args:
    focal: A tensor of shape `[A1, ..., An, 2]`, where the last dimension
      represents a camera focal length.
    principal_point: A tensor of shape `[A1, ..., An, 2]`, where the last
      dimension represents a camera principal point.
    name: A name for this op that defaults to
      "perspective_matrix_from_intrinsics".

  Returns:
    A tensor of shape `[A1, ..., An, 3, 3]`, where the last two dimensions
    represent a camera calibration matrix.

  Raises:
    ValueError: If the shape of `focal`, or `principal_point` is not
    supported.
  """
  with tf.compat.v1.name_scope(name, "perspective_matrix_from_intrinsics",
                               [focal, principal_point]):
    focal = tf.convert_to_tensor(value=focal)
    principal_point = tf.convert_to_tensor(value=principal_point)

    shape.check_static(
        tensor=focal, tensor_name="focal", has_dim_equals=(-1, 2))
    shape.check_static(
        tensor=principal_point,
        tensor_name="principal_point",
        has_dim_equals=(-1, 2))
    shape.compare_batch_dimensions(
        tensors=(focal, principal_point),
        tensor_names=("focal", "principal_point"),
        last_axes=-2,
        broadcast_compatible=False)

    fx, fy = tf.unstack(focal, axis=-1)
    cx, cy = tf.unstack(principal_point, axis=-1)
    zero = tf.zeros_like(fx)
    one = tf.ones_like(fx)
    matrix = tf.stack((fx, zero, cx,
                       zero, fy, cy,
                       zero, zero, one),
                      axis=-1)  # pyformat: disable
    matrix_shape = tf.shape(input=matrix)
    output_shape = tf.concat((matrix_shape[:-1], (3, 3)), axis=-1)
    return tf.reshape(matrix, shape=output_shape)


def project(point_3d, focal, principal_point, name=None):
  r"""Projects a 3d point onto the 2d camera plane.

  Projects a 3d point \\((x, y, z)\\) to a 2d point \\((x', y')\\) onto the
  image plane with

  $$
  \begin{matrix}
  x' = \frac{f_x}{z}x + c_x, & y' = \frac{f_y}{z}y + c_y,
  \end{matrix}
  $$

  where \\((f_x, f_y)\\) is the focal length and \\((c_x, c_y)\\) the principal
  point.

  Note:
    In the following, A1 to An are optional batch dimensions that must be
    broadcast compatible.

  Args:
    point_3d: A tensor of shape `[A1, ..., An, 3]`, where the last dimension
      represents a 3d point to project.
    focal: A tensor of shape `[A1, ..., An, 2]`, where the last dimension
      represents a camera focal length.
    principal_point: A tensor of shape `[A1, ..., An, 2]`, where the last
      dimension represents a camera principal point.
    name: A name for this op that defaults to "perspective_project".

  Returns:
    A tensor of shape `[A1, ..., An, 2]`, where the last dimension represents
    a 2d point.

  Raises:
    ValueError: If the shape of `point_3d`, `focal`, or `principal_point` is not
    supported.
  """
  with tf.compat.v1.name_scope(name, "perspective_project",
                               [point_3d, focal, principal_point]):
    point_3d = tf.convert_to_tensor(value=point_3d)
    focal = tf.convert_to_tensor(value=focal)
    principal_point = tf.convert_to_tensor(value=principal_point)

    shape.check_static(
        tensor=point_3d, tensor_name="point_3d", has_dim_equals=(-1, 3))
    shape.check_static(
        tensor=focal, tensor_name="focal", has_dim_equals=(-1, 2))
    shape.check_static(
        tensor=principal_point,
        tensor_name="principal_point",
        has_dim_equals=(-1, 2))
    shape.compare_batch_dimensions(
        tensors=(point_3d, focal, principal_point),
        tensor_names=("point_3d", "focal", "principal_point"),
        last_axes=-2,
        broadcast_compatible=True)

    point_2d, depth = tf.split(point_3d, (2, 1), axis=-1)
    point_2d *= safe_ops.safe_signed_div(focal, depth)
    point_2d += principal_point
  return point_2d


def ray(point_2d, focal, principal_point, name=None):
  r"""Computes the 3d ray for a 2d point (the z component of the ray is 1).

  Computes the 3d ray \\((r_x, r_y, 1)\\) from the camera center to a 2d point
  \\((x', y')\\) on the image plane with

  $$
  \begin{matrix}
  r_x = \frac{(x' - c_x)}{f_x}, & r_y = \frac{(y' - c_y)}{f_y}, & z = 1,
  \end{matrix}
  $$

  where \\((f_x, f_y)\\) is the focal length and \\((c_x, c_y)\\) the principal
  point. The camera optical center is assumed to be at \\((0, 0, 0)\\).

  Note:
    In the following, A1 to An are optional batch dimensions that must be
    broadcast compatible.

  Args:
    point_2d: A tensor of shape `[A1, ..., An, 2]`, where the last dimension
      represents a 2d point.
    focal: A tensor of shape `[A1, ..., An, 2]`, where the last dimension
      represents a camera focal length.
    principal_point: A tensor of shape `[A1, ..., An, 2]`, where the last
      dimension represents a camera principal point.
    name: A name for this op that defaults to "perspective_ray".

  Returns:
    A tensor of shape `[A1, ..., An, 3]`, where the last dimension represents
    a 3d ray.

  Raises:
    ValueError: If the shape of `point_2d`, `focal`, or `principal_point` is not
    supported.
  """
  with tf.compat.v1.name_scope(name, "perspective_ray",
                               [point_2d, focal, principal_point]):
    point_2d = tf.convert_to_tensor(value=point_2d)
    focal = tf.convert_to_tensor(value=focal)
    principal_point = tf.convert_to_tensor(value=principal_point)

    shape.check_static(
        tensor=point_2d, tensor_name="point_2d", has_dim_equals=(-1, 2))
    shape.check_static(
        tensor=focal, tensor_name="focal", has_dim_equals=(-1, 2))
    shape.check_static(
        tensor=principal_point,
        tensor_name="principal_point",
        has_dim_equals=(-1, 2))
    shape.compare_batch_dimensions(
        tensors=(point_2d, focal, principal_point),
        tensor_names=("point_2d", "focal", "principal_point"),
        last_axes=-2,
        broadcast_compatible=True)

    point_2d -= principal_point
    point_2d = safe_ops.safe_signed_div(point_2d, focal)
    padding = [[0, 0] for _ in point_2d.shape]
    padding[-1][-1] = 1
    return tf.pad(
        tensor=point_2d, paddings=padding, mode="CONSTANT", constant_values=1.0)


def unproject(point_2d, depth, focal, principal_point, name=None):
  r"""Unprojects a 2d point in 3d.

  Unprojects a 2d point \\((x', y')\\) to a 3d point \\((x, y, z)\\) knowing the
  depth \\(z\\) with

  $$
  \begin{matrix}
  x = \frac{z (x' - c_x)}{f_x}, & y = \frac{z(y' - c_y)}{f_y}, & z = z,
  \end{matrix}
  $$

  where \\((f_x, f_y)\\) is the focal length and \\((c_x, c_y)\\) the principal
  point.

  Note:
    In the following, A1 to An are optional batch dimensions.

  Args:
    point_2d: A tensor of shape `[A1, ..., An, 2]`, where the last dimension
      represents a 2d point to unproject.
    depth: A tensor of shape `[A1, ..., An, 1]`, where the last dimension
      represents the depth of a 2d point.
    focal: A tensor of shape `[A1, ..., An, 2]`, where the last dimension
      represents a camera focal length.
    principal_point: A tensor of shape `[A1, ..., An, 2]`, where the last
      dimension represents a camera principal point.
    name: A name for this op that defaults to "perspective_unproject".

  Returns:
    A tensor of shape `[A1, ..., An, 3]`, where the last dimension represents
    a 3d point.

  Raises:
    ValueError: If the shape of `point_2d`, `depth`, `focal`, or
    `principal_point` is not supported.
  """
  with tf.compat.v1.name_scope(name, "perspective_unproject",
                               [point_2d, depth, focal, principal_point]):
    point_2d = tf.convert_to_tensor(value=point_2d)
    depth = tf.convert_to_tensor(value=depth)
    focal = tf.convert_to_tensor(value=focal)
    principal_point = tf.convert_to_tensor(value=principal_point)

    shape.check_static(
        tensor=point_2d, tensor_name="point_2d", has_dim_equals=(-1, 2))
    shape.check_static(
        tensor=depth, tensor_name="depth", has_dim_equals=(-1, 1))
    shape.check_static(
        tensor=focal, tensor_name="focal", has_dim_equals=(-1, 2))
    shape.check_static(
        tensor=principal_point,
        tensor_name="principal_point",
        has_dim_equals=(-1, 2))
    shape.compare_batch_dimensions(
        tensors=(point_2d, depth, focal, principal_point),
        tensor_names=("point_2d", "depth", "focal", "principal_point"),
        last_axes=-2,
        broadcast_compatible=False)

    point_2d -= principal_point
    point_2d *= safe_ops.safe_signed_div(depth, focal)
    return tf.concat((point_2d, depth), axis=-1)


# API contains all public functions and classes.
__all__ = export_api.get_functions_and_classes()



# Copyright 2020 The TensorFlow Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Three.js vizualization functions."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import sys

# pylint: disable=g-import-not-at-top
try:
  from google.colab.output import _js_builder
  from google.colab.output import _publish
  from google.colab.output import eval_js
except ImportError:
  print(
      'Warning: To use the threejs_vizualization, please install the colabtools'
      ' package following the instructions detailed in the README at'
      ' github.com/tensorflow/graphics.',
      file=sys.stderr)
# pylint: enable=g-import-not-at-top


def _triangular_mesh_to_three_geometry(vertices, faces, vertex_colors=None):
  """Converts a triangular mesh to a Three.js BufferGeometry object.

  Args:
    vertices: a [V,3] numpy ndarray containing the position of the vertices
      composing the mesh. V denotes the number of vertices. Vertex positions are
      expected to be floating values.
    faces: a [F,3] numpy array describing the vertices contained in each face of
      the mesh. F denotes the number of faces. The values of that array are
      expected to be positive integer values.
    vertex_colors: a [V, 3] numpy array describing the RGB color of each vertex
      in the mesh. V denotes the number of vertices. Each channel in vertex
      colors is expected to be floating values in range [0, 1].

  Returns:
    A BufferGeometry object describing the geometry of the mesh and which can
    be consumed by Three.js.
  """
  context = _js_builder.Js(mode=_js_builder.PERSISTENT)
  vertices = context.Float32Array.new_object(vertices.ravel().tolist())
  faces = context.Uint32Array.new_object(faces.ravel().tolist())
  geometry = context.THREE.BufferGeometry.new_object()
  geometry.addAttribute('position',
                        context.THREE.BufferAttribute.new_object(vertices, 3))
  geometry.setIndex(context.THREE.BufferAttribute.new_object(faces, 1))
  geometry.computeVertexNormals()
  if vertex_colors is not None:
    vertex_colors = context.Float32Array.new_object(
        vertex_colors.ravel().tolist())
    geometry.addAttribute(
        'color', context.THREE.BufferAttribute.new_object(vertex_colors, 3))

  return geometry


def build_context():
  """Builds a javascript context."""
  threejs_url = 'https://www.gstatic.com/external_hosted/threejs-r98/'
  _publish.javascript(url=threejs_url + 'three.min.js')
  _publish.javascript(url=threejs_url + 'examples/js/controls/OrbitControls.js')
  return _js_builder.Js(mode=_js_builder.PERSISTENT)


def build_perspective_camera(field_of_view=60.0,
                             aspect_ratio=1.0,
                             near_plane=0.01,
                             far_plane=1000.0,
                             position=(0.0, 0.0, 5.0),
                             enable_zoom=False):
  """Builds a perspective camera.

  Args:
    field_of_view: the camera frustum vertical field of view.
    aspect_ratio: the camera frustum aspect ratio.
    near_plane: the camera frustum near plane.
    far_plane: the camera frustum far plane.
    position: the camera position.
    enable_zoom: whether to add a zoom functionalities to the camera controls.

  Returns:
    A Threejs camera with orbit controls.
  """
  context = build_context()
  camera = context.THREE.PerspectiveCamera.new_object(field_of_view,
                                                      aspect_ratio, near_plane,
                                                      far_plane)
  camera.position.set(*position)
  controls = context.THREE.OrbitControls.new_object(camera)
  controls.enableZoom = enable_zoom
  return camera


def triangular_mesh_renderer(meshes,
                             lights=None,
                             camera=None,
                             width=800,
                             height=600,
                             clear_color='rgb(0, 0, 0)'):
  """Function for simple mesh visualization using Three.js.

  Args:
    meshes: a list or tuple of dictionaries, each containing two required keys:
      'vertices', a [V,3] numpy ndarray of vertices, and 'faces' a [F,3] numpy
      ndarray of vertex indices that belong to each face of the mesh. V and F
      respectively correspond to the number of vertices and faces in any given
      mesh. In addition, following optional keys may be provided for each mesh:
        'vertex_colors', a [V, 3] numpy ndarray of vertex colors, and
        'material', a Three.js Material object. If a material is not provided,
        then a default Lambertian material is used.
    lights: a list of Three.js lights to add to the scene. If no light is
      provided, a point light is added to the scene at the position (0., 0.,
      5.).
    camera: a Three.js camera to add to the scene. If no camera is provided, a
      perspetive camera with orbit controls is added to the scene.
    width: the width of the rendering window.
    height: the height of the rendering window.
    clear_color: the color of the background of the rendering window.

  Returns:
    A list of Three.js BufferGeometry objects that can be used to manipulate the
    vertices of the meshes being rendered.
  """
  if not isinstance(meshes, (tuple, list)):
    meshes = [meshes]

  context = build_context()

  # Instantiate the renderer.
  renderer = context.THREE.WebGLRenderer.new_object({'antialias': True})
  renderer.setSize(width, height)
  renderer.setClearColor(context.THREE.Color.new_object(clear_color))
  renderer.clear()
  context.document.body.appendChild(renderer.domElement)

  # If no camera is supplied, create a default one.
  if camera is None:
    camera = context.THREE.PerspectiveCamera.new_object(
        75.,
        float(width) / float(height), .1, 1000.)
    camera.position.z = 5.
    controls = context.THREE.OrbitControls.new_object(camera)
    controls.enableZoom = False

  # Create the scene.
  scene = context.THREE.Scene.new_object()

  # Add lights to the scene.
  if lights is None:
    ambient_light = context.THREE.AmbientLight.new_object(0x404040)
    point_light = context.THREE.PointLight.new_object(0xffffff)
    point_light.position.set(0., 0., 5.)
    lights = (ambient_light, point_light)
  for light in lights:
    scene.add(light)

  # Build meshes and add then to the scene.
  default_material = context.THREE.MeshLambertMaterial.new_object({
      'color': 0x808080,
  })
  geometries = []
  for mesh in meshes:
    if 'vertex_colors' in mesh.keys() and mesh['vertex_colors'] is not None:
      vertex_colors = mesh['vertex_colors']
      default_material['vertexColors'] = context.THREE.VertexColors
    else:
      vertex_colors = None
      default_material['vertexColors'] = context.THREE.NoColors
    geometry = _triangular_mesh_to_three_geometry(mesh['vertices'],
                                                  mesh['faces'], vertex_colors)
    geometries.append(geometry)
    if 'material' in mesh.keys() and mesh['material'] is not None:
      scene.add(context.THREE.Mesh.new_object(geometry, mesh['material']))
    else:
      scene.add(context.THREE.Mesh.new_object(geometry, default_material))

  # Create animate function and call it to start rendering.
  rendering_loop_txt = """
			var animate = function () {
				requestAnimationFrame(animate);
				%s.render(%s, %s);
			};
			animate();
  """ % (renderer._js_value(), scene._js_value(), camera._js_value())  # pylint: disable=protected-access
  eval_js(rendering_loop_txt)
  return geometries



# Copyright 2020 The TensorFlow Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""This module implements the emission absorption voxel rendering."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import tensorflow as tf

from tensorflow_graphics.util import export_api
from tensorflow_graphics.util import shape


def render(voxels, absorption_factor=0.1, cell_size=1.0, axis=2, name=None):
  """Renders a voxel grid using the emission-absorption model, as described in ["Escaping Plato's Cave: 3D Shape From Adversarial Rendering" (Henzler 2019)](https://github.com/henzler/platonicgan).

  Note:
    In the following, A1 to An are optional batch dimensions.

  Args:
    voxels: A tensor of shape `[A1, ..., An, Vx, Vy, Vz, Vd]`, where Vx, Vy, Vz
      are the dimensions of the voxel grid and Vd the dimension of the
      information stored in each voxel (e.g. 3 for RGB color).
    absorption_factor: A scalar representing the density of the volume.
    cell_size: A scalar representing the size of a cell.
    axis: An index to the projection axis (0 for X, 1 for Y or 2 for Z).
    name: A name for this op. Defaults to "emission_absorption_render".

  Returns:
    A tensor of shape `[A1, ..., An, Vx, Vy, Vd]` representing images of size
    (Vx,Vy).

  Raises:
    ValueError: If the shape of the input tensors are not supported.
  """
  with tf.compat.v1.name_scope(name, "emission_absorption_render", [voxels]):
    voxels = tf.convert_to_tensor(value=voxels)

    shape.check_static(
        tensor=voxels, tensor_name="voxels", has_rank_greater_than=3)
    if axis not in [0, 1, 2]:
      raise ValueError("'axis' needs to be 0, 1 or 2")

    signal, density = tf.split(voxels, (-1, 1), axis=-1)

    density = tf.scalar_mul(absorption_factor / cell_size, density)
    one_minus_density = tf.ones_like(density) - density
    transmission = tf.math.cumprod(one_minus_density, axis=axis - 4)

    weight = density * transmission
    weight_sum = tf.reduce_sum(input_tensor=weight, axis=axis - 4)

    rendering = tf.reduce_sum(input_tensor=weight * signal, axis=axis - 4)
    rendering = rendering / (weight_sum + 1e-8)

    transparency = tf.reduce_prod(input_tensor=one_minus_density, axis=axis - 4)
    alpha = tf.ones_like(transparency) - transparency

    rendering = rendering * alpha

    image = tf.concat([rendering, alpha], axis=-1)
    return image


# API contains all public functions and classes.
__all__ = export_api.get_functions_and_classes()
